@@ -1143,31 +1143,32 @@ tp_notify_clickpadbutton(struct tp_dispatch *tp,
 			 enum libinput_button_state state)
 {
 	/* If we've a trackpoint, send top buttons through the trackpoint */
-	if (is_topbutton && tp->buttons.trackpoint) {
-		struct evdev_dispatch *dispatch = tp->buttons.trackpoint->dispatch;
-		struct input_event event;
-		struct input_event syn_report = {{ 0, 0 }, EV_SYN, SYN_REPORT, 0 };
+	if (tp->buttons.trackpoint) {
+		if (is_topbutton) {
+			struct evdev_dispatch *dispatch = tp->buttons.trackpoint->dispatch;
+			struct input_event event;
+			struct input_event syn_report = {{ 0, 0 }, EV_SYN, SYN_REPORT, 0 };
 
-		event.time = us2tv(time);
-		event.type = EV_KEY;
-		event.code = button;
-		event.value = (state == LIBINPUT_BUTTON_STATE_PRESSED) ? 1 : 0;
-		syn_report.time = event.time;
-		dispatch->interface->process(dispatch,
-					     tp->buttons.trackpoint,
-					     &event,
-					     time);
-		dispatch->interface->process(dispatch,
-					     tp->buttons.trackpoint,
-					     &syn_report,
-					     time);
-		return 1;
+			event.time = us2tv(time);
+			event.type = EV_KEY;
+			event.code = button;
+			event.value = (state == LIBINPUT_BUTTON_STATE_PRESSED) ? 1 : 0;
+			syn_report.time = event.time;
+			dispatch->interface->process(dispatch,
+						     tp->buttons.trackpoint,
+						     &event,
+						     time);
+			dispatch->interface->process(dispatch,
+						     tp->buttons.trackpoint,
+						     &syn_report,
+						     time);
+			return 1;
+		}
+		/* Ignore button events not for the trackpoint while suspended */
+		if (tp->device->is_suspended)
+			return 0;
 	}
 
-	/* Ignore button events not for the trackpoint while suspended */
-	if (tp->device->is_suspended)
-		return 0;
-
 	/* A button click always terminates edge scrolling, even if we
 	 * don't end up sending a button event. */
 	tp_edge_scroll_stop_events(tp, time);
